"""
The arcgis.geocoding module provides types and functions for geocoding, batch geocoding and reverse geocoding.

Geocoders can find point locations of addresses, business names, and so on.
The output points can be visualized on a map, inserted as stops for a route,
or loaded as input for spatial analysis. It is also used to generate
batch results for a set of addresses, as well as for reverse geocoding,
i.e. determining the address at a particular x/y location.
"""

from ._functions import *
