reference = {
    'journal' : {
        "values":{
            "settings":{"appGeocoders":[{"singleLineFieldName":"SingleLine",
                                         "name":"Esri World Geocoder",
                                         "url":"https://geocode.arcgis.com/arcgis/rest/services/World/GeocodeServer"},
                                        {"singleLineFieldName":"SingleLine",
                                         "name":"san",
                                         "placeholder":"",
                                         "url":"https://sampleserver5.arcgisonline.com/arcgis/rest/services/Locators/SanDiego/GeocodeServer"}
                                        ],
                        "layout":{"id":"side"}},
            "title": "title",
            "story" : {
                "storage": "WEBAPP",
                "sections": []
                },
            "template":{"name":"Map Journal",
                        "createdWith":"1.14.1",
                        "editedWith":"1.14.1"}
        }
    }
}
